@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Engine.
 # Spine Engine is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -18,3 +18,7 @@ Contains Engine's exceptions.
 
 class EngineInitFailed(Exception):
     """Raised when :class:`SpineEngine` initialization fails."""
+
+
+class RemoteEngineInitFailed(Exception):
+    """Raised when initializing the remote server connection fails."""
