@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Engine.
 # Spine Engine is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -85,21 +85,22 @@ class MultithreadExecutor(Executor):
                 waiting = {}
                 iterating = {}
                 iterating_active = set()
-                solid_names_by_jump = pipeline_context.pipeline.get_definition().loops
+                iterating_failed = set()
+                jumps = pipeline_context.pipeline.get_definition().jumps
                 jump_by_source = {}
                 jump_by_solid_name = {}
-                for jump, solid_names in solid_names_by_jump.items():
-                    jump_by_source[jump.source] = jump
-                    non_nested_solid_names = set(solid_names)
-                    for other_jump, other_solid_names in solid_names_by_jump.items():
+                for jump in jumps:
+                    jump_by_source[jump.source_solid] = jump
+                    non_nested_solid_names = set(jump.solid_names)
+                    for other_jump in jumps:
                         if jump is other_jump:
                             continue
-                        if other_solid_names > solid_names:
+                        if other_jump.solid_names > jump.solid_names:
                             continue
-                        non_nested_solid_names -= other_solid_names
+                        non_nested_solid_names -= other_jump.solid_names
                     for solid_name in non_nested_solid_names:
                         jump_by_solid_name[solid_name] = jump
-                unfinished_jumps = set(solid_names_by_jump)
+                unfinished_jumps = set(jumps)
                 loop_iteration_counters = {}
                 steps_by_key = {}
                 while not active_execution.is_complete or active_iters:
@@ -114,7 +115,7 @@ class MultithreadExecutor(Executor):
                         iterating_skipped = set()
                         for key, step in iterating.items():
                             dependency_keys = step.get_execution_dependency_keys()
-                            if dependency_keys & (iterating_active | iterating_skipped):
+                            if dependency_keys & (iterating_active | iterating_skipped | iterating_failed):
                                 iterating_skipped.add(key)
                                 continue
                             iterating_active.add(key)
@@ -123,17 +124,15 @@ class MultithreadExecutor(Executor):
                         executable_steps = []
                         for step in candidate_steps:
                             # Check if the step depends on any jumps that don't contain it
-                            jumps = (
-                                jump
-                                for jump in unfinished_jumps
-                                if step.solid_name not in solid_names_by_jump.get(jump, ())
+                            predecessor_jumps = (
+                                jump for jump in unfinished_jumps if step.solid_name not in jump.solid_names
                             )
-                            solid_names = {item for jump in jumps for item in solid_names_by_jump[jump]}
-                            problematic_keys = {
-                                key for key, step in steps_by_key.items() if step.solid_name in solid_names
+                            predecessor_solid_names = {item for jump in predecessor_jumps for item in jump.solid_names}
+                            predecessor_keys = {
+                                key for key, step in steps_by_key.items() if step.solid_name in predecessor_solid_names
                             }
                             dependency_keys = step.get_execution_dependency_keys()
-                            if dependency_keys & problematic_keys:
+                            if dependency_keys & predecessor_keys:
                                 if step.key not in iterating:
                                     waiting[step.key] = step
                                 continue
@@ -165,7 +164,7 @@ class MultithreadExecutor(Executor):
                             try:
                                 active_execution.handle_event(event_or_none)
                             except check.CheckError:
-                                # Bypass check erros on iterating steps
+                                # Bypass check errors on iterating steps
                                 if key in iterating_active:
                                     pass
                                 else:
@@ -178,11 +177,12 @@ class MultithreadExecutor(Executor):
                                 failed_jump = jump_by_solid_name.get(step.solid_name)
                                 if failed_jump is None:
                                     continue
-                                failed_solid_names = solid_names_by_jump[failed_jump]
-                                for jump, solid_names in solid_names_by_jump.items():
-                                    if solid_names & failed_solid_names:
+                                failed_solid_names = failed_jump.solid_names
+                                for jump in jumps:
+                                    if jump.solid_names & failed_solid_names:
                                         unfinished_jumps.discard(jump)
                                         loop_iteration_counters.pop(jump, None)
+                                iterating_failed.add(step)
                             elif event_or_none.is_step_success:
                                 # Process loop condition
                                 iterating_active.discard(key)
@@ -190,18 +190,18 @@ class MultithreadExecutor(Executor):
                                 jump = jump_by_source.get(step.solid_name)
                                 if jump is None:
                                     continue
-                                forward_resources = self._forward_resources.get(jump.source, [])
-                                backward_resources = self._backward_resources.get(jump.destination, [])
+                                forward_resources = self._forward_resources.get(jump.source_solid, [])
+                                backward_resources = self._backward_resources.get(jump.destination_solid, [])
                                 jump.receive_resources_from_source(forward_resources)
                                 jump.receive_resources_from_destination(backward_resources)
                                 iteration_counter = loop_iteration_counters.setdefault(jump, 1)
                                 if jump.is_condition_true(iteration_counter):
                                     # Put all jump steps in the iterating bucket
                                     for k, s in steps_by_key.items():
-                                        if s.solid_name in solid_names_by_jump[jump]:
+                                        if s.solid_name in jump.solid_names:
                                             iterating[k] = s
                                     # Mark all nested jumps unfinished again
-                                    for solid_name in solid_names_by_jump[jump]:
+                                    for solid_name in jump.solid_names:
                                         nested_jump = jump_by_solid_name.get(solid_name)
                                         if nested_jump is not None:
                                             unfinished_jumps.add(nested_jump)
