@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Engine.
 # Spine Engine is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -16,12 +16,15 @@ Provides the ProjectItemResource class.
 :date:   29.4.2020
 """
 import copy
+import uuid
 from contextlib import contextmanager
 from pathlib import Path
 from urllib.parse import urlparse
 from urllib.request import url2pathname
 from spinedb_api.filters.tools import clear_filter_configs
-from spinedb_api.spine_db_server import closing_spine_db_server
+from spinedb_api.spine_db_server import closing_spine_db_server, quick_db_checkout
+from spinedb_api.spine_db_client import SpineDBClient
+from ..utils.helpers import PartCount
 
 
 class ProjectItemResource:
@@ -34,7 +37,7 @@ class ProjectItemResource:
         metadata (dict): resource's metadata
     """
 
-    def __init__(self, provider_name, type_, label, url=None, metadata=None):
+    def __init__(self, provider_name, type_, label, url=None, metadata=None, filterable=False, identifier=None):
         """
         Args:
             provider_name (str): The name of the item that provides the resource
@@ -50,6 +53,8 @@ class ProjectItemResource:
 
                 - filter_stack (str): resource's filter stack
                 - filter_id (str): filter id
+            filterable (bool): If True, the resource provides opportunity for filtering
+            identifier (str): an identifier of the original instance, shared also by all the clones
         """
         self.provider_name = provider_name
         self.type_ = type_
@@ -57,17 +62,44 @@ class ProjectItemResource:
         self._url = url
         self._parsed_url = urlparse(self._url)
         self.metadata = metadata if metadata is not None else dict()
+        self._filterable = filterable
+        self._identifier = identifier if identifier is not None else uuid.uuid4().hex
 
     @contextmanager
-    def open(self):
+    def open(self, db_checkin=False, db_checkout=False):
         if self.type_ == "database":
-            with closing_spine_db_server(self.url, memory=self.metadata.get("memory", False)) as server_url:
-                yield server_url
+            ordering = {
+                "id": self._identifier,
+                "part_count": self.metadata.get("part_count", PartCount()),
+                "current": self.metadata.get("current"),
+                "precursors": self.metadata.get("precursors", set()),
+            }
+            with closing_spine_db_server(
+                self.url, memory=self.metadata.get("memory", False), ordering=ordering
+            ) as server_url:
+                if db_checkin:
+                    SpineDBClient.from_server_url(server_url).db_checkin()
+                try:
+                    yield server_url
+                finally:
+                    if db_checkout:
+                        SpineDBClient.from_server_url(server_url).db_checkout()
         else:
             yield self.path if self.hasfilepath else ""
 
+    def quick_db_checkout(self):
+        if self.type_ != "database":
+            return
+        ordering = {
+            "id": self._identifier,
+            "part_count": self.metadata.get("part_count", PartCount()),
+            "current": self.metadata.get("current"),
+            "precursors": self.metadata.get("precursors", set()),
+        }
+        quick_db_checkout(ordering)
+
     def clone(self, additional_metadata=None):
-        """Clones a resource and optionally updates the clone's metadata.
+        """Clones this resource and optionally updates the clone's metadata.
 
         Args:
             additional_metadata (dict): metadata to add to the clone
@@ -79,7 +111,15 @@ class ProjectItemResource:
             additional_metadata = {}
         metadata = copy.deepcopy(self.metadata)
         metadata.update(additional_metadata)
-        return ProjectItemResource(self.provider_name, self.type_, label=self.label, url=self._url, metadata=metadata)
+        return ProjectItemResource(
+            self.provider_name,
+            self.type_,
+            label=self.label,
+            url=self._url,
+            metadata=metadata,
+            filterable=self._filterable,
+            identifier=self._identifier,
+        )
 
     def __eq__(self, other):
         if not isinstance(other, ProjectItemResource):
@@ -90,6 +130,7 @@ class ProjectItemResource:
             and self.type_ == other.type_
             and self._url == other._url
             and self.metadata == other.metadata
+            and self._filterable == other._filterable
         )
 
     def __hash__(self):
@@ -100,7 +141,8 @@ class ProjectItemResource:
         result += f"provider={self.provider_name}, "
         result += f"type_={self.type_}, "
         result += f"url={self._url}, "
-        result += f"metadata={self.metadata})"
+        result += f"metadata={self.metadata}, "
+        result += f"filterable={self._filterable})"
         return result
 
     @property
@@ -133,6 +175,10 @@ class ProjectItemResource:
     def arg(self):
         return self._url if self.type_ == "database" else self.path
 
+    @property
+    def filterable(self):
+        return self._filterable
+
 
 class CmdLineArg:
     """Command line argument for items that execute shell commands."""
@@ -170,7 +216,7 @@ class LabelArg(CmdLineArg):
         return {"type": "resource", "arg": self.arg}
 
 
-def database_resource(provider_name, url, label=None):
+def database_resource(provider_name, url, label=None, filterable=False):
     """
     Constructs a database resource.
 
@@ -178,10 +224,11 @@ def database_resource(provider_name, url, label=None):
         provider_name (str): resource provider's name
         url (str): database URL
         label (str, optional): resource label
+        filterable (bool): is resource filterable
     """
     if label is None:
         label = clear_filter_configs(url)
-    return ProjectItemResource(provider_name, "database", label, url)
+    return ProjectItemResource(provider_name, "database", label, url, filterable=filterable)
 
 
 def file_resource(provider_name, file_path, label=None):
@@ -275,21 +322,23 @@ def get_labelled_sources(resources):
     return d
 
 
-def cmd_line_arg_from_dict(arg_dict):
+def make_cmd_line_arg(arg_spec):
     """Deserializes argument from dictionary.
 
     Args:
-        arg_dict (dict): serialized command line argument
+        arg_spec (dict or str): serialized command line argument
 
     Returns:
         CmdLineArg: deserialized command line argument
     """
-    type_ = arg_dict["type"]
+    if not isinstance(arg_spec, dict):
+        return CmdLineArg(arg_spec)
+    type_ = arg_spec["type"]
     construct = {"literal": CmdLineArg, "resource": LabelArg}[type_]
-    return construct(arg_dict["arg"])
+    return construct(arg_spec["arg"])
 
 
-def labelled_resource_args(resources, stack):
+def labelled_resource_args(resources, stack, db_checkin=False, db_checkout=False):
     """
     Args:
         resources (Iterable of ProjectItemResource): resources to process
@@ -301,9 +350,11 @@ def labelled_resource_args(resources, stack):
     result = {}
     single_resources, pack_resources = extract_packs(resources)
     for resource in single_resources:
-        result[resource.label] = stack.enter_context(resource.open())
+        result[resource.label] = stack.enter_context(resource.open(db_checkin=db_checkin, db_checkout=db_checkout))
     for label, resources_ in pack_resources.items():
-        result[label] = " ".join(stack.enter_context(r.open()) for r in resources_)
+        result[label] = " ".join(
+            stack.enter_context(r.open(db_checkin=db_checkin, db_checkout=db_checkout)) for r in resources_
+        )
     return result
 
 
