@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Engine.
 # Spine Engine is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -20,14 +20,36 @@ import sys
 import datetime
 import time
 import json
-from urllib.parse import urlparse, urlunparse
 from pathlib import Path
-
+from enum import Enum, auto, unique
 import networkx
 from jupyter_client.kernelspec import find_kernel_specs
 from ..config import PYTHON_EXECUTABLE, JULIA_EXECUTABLE, GAMS_EXECUTABLE, EMBEDDED_PYTHON
 
 
+@unique
+class ExecutionDirection(Enum):
+    FORWARD = auto()
+    BACKWARD = auto()
+    NONE = auto()
+
+    def __str__(self):
+        return str(self.name)
+
+
+@unique
+class ItemExecutionFinishState(Enum):
+    SUCCESS = 1
+    FAILURE = 2
+    SKIPPED = 3
+    EXCLUDED = 4
+    STOPPED = 5
+    NEVER_FINISHED = 6
+
+    def __str__(self):
+        return str(self.name)
+
+
 class Singleton(type):
     _instances = {}
 
@@ -247,21 +269,6 @@ def write_filter_id_file(filter_id, path):
         filter_id_file.writelines([filter_id + "\n"])
 
 
-def remove_credentials_from_url(url):
-    """Removes username and password information from URLs.
-
-    Args:
-        url (str): URL
-
-    Returns:
-        str: sanitized URL
-    """
-    parsed = urlparse(url)
-    if parsed.username is None:
-        return url
-    return urlunparse(parsed._replace(netloc=parsed.netloc.partition("@")[-1]))
-
-
 def gather_leaf_data(input_dict, paths, pop=False):
     """Gathers data defined by 'paths' of keys from nested dicts.
 
@@ -300,3 +307,48 @@ def gather_leaf_data(input_dict, paths, pop=False):
         leaf_dict = build_to_leaf(output_dict, prefix)
         leaf_dict[prefix[-1]] = value
     return output_dict
+
+
+def get_file_size(size_in_bytes):
+    """Returns a human readable string of the size of a file. Given size_in_bytes arg
+    is designed as the output of os.path.getsize().
+
+    1 KB = 1024 bytes
+    1 MB = 1024*1024 bytes
+    1 GB = 1024*1024*1024 bytes
+
+    Args:
+        size_in_bytes (int): Size in bytes [B]
+
+    Returns:
+        str: Human readable file size
+    """
+    kb = 1024
+    mb = 1024 * 1024
+    gb = 1024 * 1024 * 1024
+    if size_in_bytes <= kb:
+        return str(size_in_bytes) + " B"
+    if kb < size_in_bytes <= mb:
+        return str(round(size_in_bytes / kb, 1)) + " KB"
+    elif mb < size_in_bytes < gb:
+        return str(round(size_in_bytes / mb, 1)) + " MB"
+    else:
+        return str(round(size_in_bytes / gb, 1)) + " GB"
+
+
+class PartCount:
+    def __init__(self):
+        self._count = 0
+
+    def __iadd__(self, number):
+        self._count += number
+        return self
+
+    def __eq__(self, number):
+        return self._count == number
+
+    def __deepcopy__(self, memo):
+        return self
+
+    def __repr__(self):
+        return str(self._count)
